import subprocess
import warnings
from functools import lru_cache
from pathlib import Path

import requests

from ak_transcribe.utils.config_parser import Config

_config = Config()


class _WhisperParser:
    def __init__(self) -> None:
        self.results: dict | None = None
        self.mp3path: Path | None = None

    def transcript(self) -> str:
        assert isinstance(self.results, dict), "Results not set yet"
        return "\n".join(
            [line["text"].strip() for line in self.results["segments"]]  # type: ignore
        )

    def srt(self) -> str:
        assert isinstance(self.results, dict), "Results not set yet"
        count = 0
        text_blob = ""
        for segment in self.results.get("segments", []):
            count += 1
            _timestamp_start: str = f"{count}\n{self._srt_time(segment['start'])}"
            _timestamp_end: str = f"{self._srt_time(segment['end'])}\n"
            _text: str = f"{segment['text'].replace('-->', '->').strip()}"
            text_blob += f"{_timestamp_start} --> {_timestamp_end}{_text}\n\n"
        return text_blob

    @staticmethod
    def _srt_time(_seconds: float) -> str:
        """Convert seconds into SRT (SubRip Text) time format."""
        assert _seconds >= 0, "non-negative timestamp expected"
        milliseconds: int = round(_seconds * 1000.0)
        hours: int = milliseconds // 3_600_000
        milliseconds -= hours * 3_600_000
        minutes: int = milliseconds // 60_000
        milliseconds -= minutes * 60_000
        seconds: int = milliseconds // 1_000
        milliseconds -= seconds * 1_000
        return (f"{hours}:") + f"{minutes:02d}:{seconds:02d},{milliseconds:03d}"


class TempMp3:
    def __init__(self, filepath: Path):
        self.unique_mp3file: bool = False
        self.mp3path: Path | None = None
        self._ffmpeg_path: Path | str = Path(
            _config.get(keys=("ffmpeg", "cmd"), default="ffmpeg")
        )
        self.filepath = filepath

    def __enter__(self) -> Path:
        self.unique_mp3file = self.filepath.suffix.casefold() != ".mp3"
        if self.unique_mp3file:
            self.mp3path = self._convert_to_audio(file=self.filepath)
        else:
            self.mp3path = self.filepath
        return self.mp3path

    def __exit__(self, exc_type, exc_value, traceback):
        if self.unique_mp3file and self.mp3path is not None:
            self.mp3path.unlink()
        if exc_type is not None:
            print(f"An exception occurred: {exc_type}, {exc_value}")
        return False

    def _convert_to_audio(self, file: Path) -> Path:
        """Convert the given audio file to an MP3 format."""
        outputpath = (
            Path(_config.get(keys=("ffmpeg", "transcode_dir"), default="."))
            / f"{file.stem}.mp3"
        )
        command = f'"{self._ffmpeg_path}" -loglevel quiet -stats -hide_banner -i "{file}" "{outputpath}" -y'
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        process.wait()
        if process.returncode != 0:
            raise Exception("Conversion to `.mp3` failed.")
        return outputpath


class _OpenAI_CPU(_WhisperParser):
    def __init__(self) -> None:
        try:
            import whisper  # type: ignore
        except ModuleNotFoundError:
            raise Exception(
                (
                    "Whisper not installed. "
                    "Install using `pip install openai-whisper@git+https://github.com/openai/whisper.git`"
                )
            )

        model = _config.get(keys=("whisper", "model"), default="small.en")
        device = _config.get(keys=("whisper", "device"), default="cpu")
        self.whisper = whisper.load_model(model, device=device)
        super().__init__()

    def process(self, filepath: Path):
        """Process the file and save results to `self.results`"""
        with TempMp3(filepath=filepath) as file:
            with warnings.catch_warnings(action="ignore"):
                self.results = self.whisper.transcribe(str(file))


class _OpenAI_webserver(_WhisperParser):
    def __init__(self) -> None:
        self.base_url: str = _config.get(keys=("whisper", "url"), default="")
        assert self.base_url != "", "Make sure [whisper][url] is set in `config.toml`"
        requests.get(self.base_url).raise_for_status()

        self.transcript_url = f"{self.base_url.rstrip('/')}/asr?encode=true&task=transcribe&word_timestamps=true&output=json"

    def process(self, filepath: Path):
        """Process the file and save results to `self.results`"""

        with TempMp3(filepath=filepath) as file:
            audio_file_path = str(file)
            _files = {
                "audio_file": (
                    audio_file_path,
                    open(audio_file_path, "rb"),
                    "audio/mpeg",
                )
            }
            response = requests.post(self.transcript_url, files=_files)
            response.raise_for_status()
            self.results = response.json()


@lru_cache(maxsize=1)
def _get_processor() -> _OpenAI_webserver | _OpenAI_CPU:
    try:
        return _OpenAI_webserver()
    except Exception as e:
        return _OpenAI_CPU()


Transcriber: _OpenAI_webserver | _OpenAI_CPU = _get_processor()
